@@ -1,5 +1,5 @@
 //
-// $Id: dnstracer.c,v 1.40 2003/02/14 09:58:54 mavetju Exp $
+// $Id: dnstracer.c,v 1.48 2004/07/08 11:15:17 mavetju Exp $
 //
 
 //
@@ -60,17 +60,21 @@
 #include "config.h"
 
 #define DEFAULT_RETRIES	3
-#define DEFAULT_TIMEOUT	15
 #define DEFAULT_CACHING	1
+#define DEFAULT_NEGATIVE_CACHING	0
 #define DEFAULT_OVERVIEW	0
 #define DEFAULT_QUERYTYPE	ns_t_a
 #define DEFAULT_NOIPV6	0
 
+#ifdef NOIPV6
+#define gethostbyname2(a,b) gethostbyname(a)
+#endif
+
 int	verbose=0;
 int	global_overview=DEFAULT_OVERVIEW;
 int	global_retries=DEFAULT_RETRIES;
-int	global_timeout=DEFAULT_TIMEOUT;
 int	global_caching=DEFAULT_CACHING;
+int	global_negative_caching=DEFAULT_NEGATIVE_CACHING;
 int	global_querytype=DEFAULT_QUERYTYPE;
 int	global_noipv6=DEFAULT_NOIPV6;
 char *	global_source_address=NULL;
@@ -172,6 +176,41 @@ struct dnsrr {
 	struct dnsrr *	next;
 };
 
+char *rr_types[256]={
+	"#0x00", "a", "ns", "#0x03", "#0x04", "cname", "soa", "#0x07",
+	"#0x08", "#0x09", "#0x0a", "#0x0b", "#ptr", "#hinfo", "#0x0e", "#mx",
+	"#txt", "#0x11", "#0x12", "#0x13", "#0x14", "#0x15", "#0x16", "#0x17",
+	"#0x18", "#0x19", "#0x1a", "#0x1b", "#0x1c", "#0x1d", "#0x1e", "#0x1f",
+	"#0x20", "#0x21", "#0x22", "#0x23", "#0x24", "#0x25", "#0x26", "#0x27",
+	"#0x28", "#0x29", "#0x2a", "#0x2b", "#0x2c", "#0x2d", "#0x2e", "#0x2f",
+	"#0x30", "#0x31", "#0x32", "#0x33", "#0x34", "#0x35", "#0x36", "#0x37",
+	"#0x38", "#0x39", "#0x3a", "#0x3b", "#0x3c", "#0x3d", "#0x3e", "#0x3f",
+	"#0x40", "#0x41", "#0x42", "#0x43", "#0x44", "#0x45", "#0x46", "#0x47",
+	"#0x48", "#0x49", "#0x4a", "#0x4b", "#0x4c", "#0x4d", "#0x4e", "#0x4f",
+	"#0x50", "#0x51", "#0x52", "#0x53", "#0x54", "#0x55", "#0x56", "#0x57",
+	"#0x58", "#0x59", "#0x5a", "#0x5b", "#0x5c", "#0x5d", "#0x5e", "#0x5f",
+	"#0x60", "#0x61", "#0x62", "#0x63", "#0x64", "#0x65", "#0x66", "#0x67",
+	"#0x68", "#0x69", "#0x6a", "#0x6b", "#0x6c", "#0x6d", "#0x6e", "#0x6f",
+	"#0x70", "#0x71", "#0x72", "#0x73", "#0x74", "#0x75", "#0x76", "#0x77",
+	"#0x78", "#0x79", "#0x7a", "#0x7b", "#0x7c", "#0x7d", "#0x7e", "#0x7f",
+	"#0x80", "#0x81", "#0x82", "#0x83", "#0x84", "#0x85", "#0x86", "#0x87",
+	"#0x88", "#0x89", "#0x8a", "#0x8b", "#0x8c", "#0x8d", "#0x8e", "#0x8f",
+	"#0x90", "#0x91", "#0x92", "#0x93", "#0x94", "#0x95", "#0x96", "#0x97",
+	"#0x98", "#0x99", "#0x9a", "#0x9b", "#0x9c", "#0x9d", "#0x9e", "#0x9f",
+	"#0xa0", "#0xa1", "#0xa2", "#0xa3", "#0xa4", "#0xa5", "#0xa6", "#0xa7",
+	"#0xa8", "#0xa9", "#0xaa", "#0xab", "#0xac", "#0xad", "#0xae", "#0xaf",
+	"#0xb0", "#0xb1", "#0xb2", "#0xb3", "#0xb4", "#0xb5", "#0xb6", "#0xb7",
+	"#0xb8", "#0xb9", "#0xba", "#0xbb", "#0xbc", "#0xbd", "#0xbe", "#0xbf",
+	"#0xc0", "#0xc1", "#0xc2", "#0xc3", "#0xc4", "#0xc5", "#0xc6", "#0xc7",
+	"#0xc8", "#0xc9", "#0xca", "#0xcb", "#0xcc", "#0xcd", "#0xce", "#0xcf",
+	"#0xd0", "#0xd1", "#0xd2", "#0xd3", "#0xd4", "#0xd5", "#0xd6", "#0xd7",
+	"#0xd8", "#0xd9", "#0xda", "#0xdb", "#0xdc", "#0xdd", "#0xde", "#0xdf",
+	"#0xe0", "#0xe1", "#0xe2", "#0xe3", "#0xe4", "#0xe5", "#0xe6", "#0xe7",
+	"#0xe8", "#0xe9", "#0xea", "#0xeb", "#0xec", "#0xed", "#0xee", "#0xef",
+	"#0xf0", "#0xf1", "#0xf2", "#0xf3", "#0xf4", "#0xf5", "#0xf6", "#0xf7",
+	"#0xf8", "#0xf9", "#0xfa", "#0xfb", "#0xfc", "#0xfd", "#0xfe", "#0xff"
+};
+
 /*****************************************************************************/
 
 char *get_resource(int type,struct dnssession *session,char *buffer,int dots);
@@ -522,13 +561,13 @@ char *get_resource(int type,struct dnssession *session,char *buffer,int dots) {
 	static char retval[3*NS_MAXDNAME];
 	char mname[NS_MAXDNAME];
 	char rname[NS_MAXDNAME];
-	unsigned short us;
+	unsigned long ul;
 
 	strcpy(mname,printablename(getname(session,&buffer),dots));
 	strcpy(rname,printablename(getname(session,&buffer),dots));
 
-	us=getshort(buffer);
-	sprintf(retval,"serial: %hu mname: %s rname: %s",us,mname,rname);
+	ul=getlong(buffer);
+	sprintf(retval,"serial: %ld mname: %s rname: %s",ul,mname,rname);
 	return retval;
     }
 	
@@ -729,7 +768,7 @@ int create_socket(int PF) {
         hints.ai_family = PF_INET;
         hints.ai_socktype = SOCK_DGRAM;
         error = getaddrinfo(global_source_address, 0, &hints, &src_res);
-        if (error == EAI_NODATA) {
+        if (error == EAI_NONAME) {
 	    hints.ai_flags = 0;
 	    error = getaddrinfo(global_source_address, 0, &hints, &src_res);
         }
@@ -796,13 +835,13 @@ int send_data(char *server,struct dnssession *session) {
     return cc;
 }
 
-int receive_data(struct dnssession *session) {
+int receive_data(struct dnssession *session,int retry) {
     char		buffer[2048];
     int			len;
     fd_set		in_set;
     struct timeval	timeout;
 
-    timeout.tv_sec=global_timeout;
+    timeout.tv_sec=5*(1<<retry);
     timeout.tv_usec=0;
 
     FD_ZERO(&in_set);
@@ -1020,7 +1059,7 @@ int create_session(char *host,
     struct dnssession * session;
     struct dnsrr *	rrauth;
     struct dnsrr *	rradd;
-    int			i,retval,errorcode;
+    int			i,retval,errorcode,refersbackwards=0;
     char		s[NS_MAXDNAME];
 
     //
@@ -1097,7 +1136,7 @@ int create_session(char *host,
     errorcode=0;
     for (i=0;i<global_retries;i++) {
 	send_data(server_ip,session);
-	if ((errorcode=receive_data(session))==0)
+	if ((errorcode=receive_data(session,i))==0)
 	    break;
 	printf("* ");
 	fflush(stdout);
@@ -1109,6 +1148,7 @@ int create_session(char *host,
     //
     if (errorcode!=0) {
 	remove_busy(server_ip);
+	if (global_negative_caching) add_answer(server_ip);
 	return 1;
     }
 
@@ -1126,6 +1166,8 @@ int create_session(char *host,
 
 	answer=session->answer;
 	while (answer!=NULL) {
+	    if (answer->type!=global_querytype)
+		printf("[received type is %s] ",rr_types[answer->type]);
 	    add_arecord(session,answer,server_name,server_ip);
 	    answer=answer->next;
 	}
@@ -1183,9 +1225,17 @@ int create_session(char *host,
 	// of what we have, don't go there. This might happen if we are
 	// looking through cnames from different domains.
 	//
-	if (strcmp(rrauth->domainname_string,".") &&
-		strcasecmp(rrauth->domainname_string,
-		    host+strlen(host)-strlen(rrauth->domainname_string))!=0) {
+	if (strcmp(rrauth->domainname_string,".")==0) {
+	    rrauth=rrauth->next;
+	    if (!refersbackwards++)
+		printf("Refers backwards ");
+	    continue;
+	}
+
+	if (server_authfor!=NULL && strcmp(server_authfor,".")!=0 &&
+	    (unsigned char *)strcasestr(rrauth->domainname_string,server_authfor)!=rrauth->domainname_string+strlen(rrauth->domainname_string)-strlen(server_authfor)) {
+	    if (!refersbackwards++)
+		printf("Refers backwards ");
 	    rrauth=rrauth->next;
 	    continue;
 	}
@@ -1233,7 +1283,7 @@ int create_session(char *host,
 		strcpy(nextserver_name,printablename(rradd->domainname,1));
 		strcpy(nextserver_ip,rradd->data_string);
 		retval+=create_session(host,
-			    nextserver_ip,rradd->type==ns_t_aaaa,
+			    nextserver_ip,(rradd->type==ns_t_aaaa)?1:0,
 			    nextserver_name,rrauth->domainname_string,
 			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
 	    } else {
@@ -1241,7 +1291,11 @@ int create_session(char *host,
 
 		strcpy(nextserver_name,rrauth->data_string);
 
+#ifdef NOIPV6
+		for (ip=0;ip<1;ip++) {
+#else
 		for (ip=0;ip<2;ip++) {
+#endif
 		    int count,i;
 		    struct hostent *h;
 		    char **addr_list=NULL;
@@ -1333,19 +1387,23 @@ int create_session(char *host,
 
 void usage(void) {
     fprintf(stderr,
-	"DNSTRACER version 1.7 - (c) Edwin Groothuis - http://www.mavetju.org\n"
+	"DNSTRACER version 1.8 - (c) Edwin Groothuis - http://www.mavetju.org\n"
 	"Usage: dnstracer [options] [host]\n"
 	"\t-c: disable local caching, default enabled\n"
+	"\t-C: enable negative caching, default disabled\n"
 	"\t-o: enable overview of received answers, default disabled\n"
 	"\t-q <querytype>: query-type to use for the DNS requests, default A\n"
 	"\t-r <retries>: amount of retries for DNS requests, default 3\n"
 	"\t-s <server>: use this server for the initial request, default localhost\n"
 	"\t             If . is specified, A.ROOT-SERVERS.NET will be used.\n"
-	"\t-t <timeout>: timeout for DNS requests, default 15 seconds\n"
 	"\t-v: verbose\n"
-	"\t-4: don't query IPv6 servers\n"
 	"\t-S <ip address>: use this source address.\n"
     );
+#ifndef NOIPV6
+    fprintf(stderr,
+	"\t-4: don't query IPv6 servers\n"
+    );
+#endif
     exit(1);
 }
 
@@ -1390,16 +1448,26 @@ int main(int argc,char **argv) {
     wsockinit();
 #endif
 
-    while ((ch=getopt(argc,argv,"4coq:r:S:s:t:v"))!=-1) {
+#ifndef NOIPV6
+    while ((ch=getopt(argc,argv,"4cCoq:r:S:s:v"))!=-1) {
+#else
+    while ((ch=getopt(argc,argv,"cCoq:r:S:s:v"))!=-1) {
+#endif
 	switch (ch) {
+#ifndef NOIPV6
 	case '4':
 	    global_noipv6=1;
 	    break;
+#endif
 
 	case 'c':
 	    global_caching=0;
 	    break;
 
+	case 'C':
+	    global_negative_caching=1;
+	    break;
+
 	case 'o':
 	    global_overview=1;
 	    break;
@@ -1448,14 +1516,6 @@ int main(int argc,char **argv) {
 	    }
 	    break;
 
-	case 't':
-	    if ((global_timeout=atoi(optarg))<1) {
-		fprintf(stderr,
-		    "Strange timeout, setting to default\n");
-		global_timeout=DEFAULT_TIMEOUT;
-	    }
-	    break;
-
 	case 'v':
 	    verbose=1;
 	    break;
@@ -1473,8 +1533,8 @@ int main(int argc,char **argv) {
     strcpy(argv0,argv[0]);
     if (argv0[strlen(argv[0])-1]=='.') argv0[strlen(argv[0])-1]=0;
 
-    printf("Tracing to %s via %s, timeout %d seconds\n",
-	argv0,server_name,global_timeout);
+    printf("Tracing to %s[%s] via %s, maximum of %d retries\n",
+	argv0,rr_types[global_querytype],server_name,global_retries);
 
     srandom(time(NULL));
 
